/*
 * ASCD : copyright (c) Aley Keprt 1998-2001
 *
 * File tape.c - tape file handling
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include "zlib.h"

#include "z80.h"
#include "tape.h"
#include "mem.h"
#include "kbd.h"


#define xDEBUG_TAPE


typedef unsigned char byte;


static gzFile f;
static int startup; //set to 1 when autoloading Sam .tap file


typedef struct {
} Z80_header;


//-----------------------------------------------------------------------------------


int open_tap(const char *fname) {
//
//open tape file
//
//return: error flag
//
  Z80_Regs r;

  close_tap();

  f=gzopen(fname,"rb");
  if(!f) return -1;

  //select emulation mode at startup
  //decide by looking at the first block type: 0=ZXS, 1=Sam
  get_registers(&r);
  if(r.pc==0) {
    int type;
    gzgetc(f);
    gzgetc(f);
    type=gzgetc(f);
    gzrewind(f);

    if(type==0 && cfg.zxmode!=EM_48k) {
      //zx128k autoload
      set_emulation_mode(EM_128k);
      keyports[6]&=~1; //Enter
      startup=1;
    } else
    if(type==1) {
      //Sam autoload
      set_emulation_mode(EM_Sam);
      keyports[2]&=~0x20; //F7
      startup=1;
    }
  }

  return 0;
}


//-----------------------------------------------------------------------


void close_tap(void) {
//
//close open tape file (if any)
//
  if(f) {gzclose(f); f=0;}
}


//-----------------------------------------------------------------------


static int save_tape(int src,int size,int type) {
//
//save to tape
//
//return: error flag
//
  FILE *f;
  int error,pos;
  int crc;

  if(size>65535-2) return -1; //loo large block to fit in .tap file!

  f=fopen(cfg.tape,"ab");
  if(!f) return -1;
  pos=ftell(f);

  //block size
  size+=2;
  fwrite(&size,2,1,f);
  size-=2;

  //type byte
  fwrite(&type,1,1,f);
  crc=type;

  //body
  for(;size--;) {
    int value=read_byte(src);
    crc^=value;
    fwrite(&value,1,1,f);

    src++;
    if(cfg.zxmode==EM_Sam) {
      //keep src in section C
      if(src>=0xc000) {out_hmpr(hmpr+1); src&=0x80ff;}
    }
  }

  //crc
  fwrite(&crc,1,1,f);

  error=ferror(f);
  fclose(f);

  if(error) {
    //first aid = remove corrupted block
    truncate(cfg.tape,pos);
  }

  return error;
}


//-----------------------------------------------------------------------


void tap_562_trap(void) {
//
//ZX-ROM 0x562 trap
//
//see mem.cpp for list of actually trapped addresses
//
  Z80_Regs r;
  int restarted=0;
  int size;
  unsigned short dest=0; //Sam-specific destination

  static unsigned short tap_size;
  static byte *data;

  if(startup) {
    //release F7 key
    startup=0;
    if(cfg.zxmode==EM_Sam) keyports[2]|=0x20; else keyports[6]|=1;
  }

again:
  get_registers(&r);
  size=r.d*256+r.e;

  //convert Sam registers to ZXS registers
  if(cfg.zxmode==EM_Sam) {
    dest=r.h*256+r.l;

    if(r.pc==0xe56a+2) {
      //Sam save
      int error=save_tape(dest,size+r.c,r.a);
      r.f=error ? r.f&~F_CARRY : r.f|F_CARRY;

      set_registers(&r);
      return;
    }

    size+=read_byte(0x5ac8)*65536;  //0x5ac8 = stored c = size/65536
  } else {
    if(r.pc==0x4d8+2) {
      //ZXS save
      int error=save_tape(r.ix+1,size-1,r.a1);
      r.f=error ? r.f&~F_CARRY : r.f|F_CARRY;

      set_registers(&r);
      return;
    }
  }

#ifdef DEBUG_TAPE
  printf("%s dest=%i, size=%i, type=%i\n",r.f1&CARRY ? "load" : "verify", r.ix,size,r.a1);
#endif

  //f1 cy=0 --> verify
  if(!(r.f1&F_CARRY)) {
    //return immediately
    //cy=0,z=0 --> indicate error
    r.f&=~(F_ZERO|F_CARRY);
    return;
  }

  if(r.pc==0x5ba+2 && data) {
    //re-entry after type byte was rejected!
    //this hack makes mr.copy+ working!
    r.a1=data[0];

#ifdef DEBUG_TAPE
    printf("re-entry!\n");
#endif

  } else {

    //read block size
    if(2!=gzread(f,&tap_size,2)) goto restart;

    //read data block (type+[body]+crc --> 2 bytes longer)
    if(data) free(data);
    data=(byte*)malloc(tap_size);
    if(tap_size != gzread(f,data,tap_size)) goto restart;
  }

#ifdef DEBUG_TAPE
  printf("found: size=%i, type=%i\n",tap_size-2,data[0]);
#endif

  //Sam specific: always accept ZXS headers (strange behaviour of Sam ROM)
  if(cfg.zxmode==EM_Sam && !(r.f1&F_ZERO) && data[0]==0) {
    r.a1=0;
    r.e=17;
    size=(size&~255)+17; //lowest byte is always set to 17!
  }

  //note: f1 zf=1 --> load type byte, and don't test it
  if((r.f1&F_ZERO) || r.a1==data[0]) {

    byte *ptr=data;
    int wanted=size; //wanted size - will be used as counter
    int crc=0;

    //skip type byte if it is not to be stored into memory
    if(!(r.f1&F_ZERO)) {
      ptr++;
      tap_size--;
      crc=*data; //first byte of crc
    } else {
#ifdef DEBUG_TAPE
      printf("ZF=1\n");
#endif
    }

    //move data to memory
    if(cfg.zxmode==EM_Sam) {

      //Sam
      for(;tap_size && wanted;tap_size--,wanted--) {
        crc^=*ptr;
        write_byte(dest,*ptr++);

        //keep dest in section C
        if(++dest>=0xc000) {out_hmpr(hmpr+1); dest&=0x80ff;}
      }

    } else {

      //ZXS
      for(;tap_size && wanted;tap_size--,wanted--) {
        crc^=*ptr;
        write_byte(r.ix++,*ptr++);
      }
    }

    //if there is still something on tape, it's crc byte
    if(tap_size) {
      //if there are more bytes, it's harmless as long as crc is valid
      crc^=*ptr; //ZXS
    }

    //update Z80 registers
    //ix=dest+size (already set)
    //h=crc (already set)
    r.d=wanted>>8;  //remaining size
    r.e=wanted;

    if(cfg.zxmode==EM_Sam) {
      r.l=crc;
      r.h=*ptr; //last byte read
    } else {
      r.h=crc;
      r.l=*ptr; //last byte read
    }

    if(crc || wanted) {
      //bad crc or size --> cy=0
      r.f&=~F_CARRY;
    } else {
      //good crc and size --> cy=1
      r.f|=F_CARRY;
    }
    r.a=r.l^r.h; //a = read^wanted (zero=crc ok)

    //free buffer
    free(data);
    data=0;

  } else {

    //type byte is different
    r.l=r.h=data[0]; //read byte and checksum (equal to each other)
    r.a=r.a1^r.l;    //read^expected
    //NZ already set
  }

  //restore registers
  set_registers(&r);
  return;

restart:

  free(data);
  data=0;

  if(!restarted) {
    gzrewind(f);
    restarted=1;
    goto again;
  }
  close_tap();

  //tape is at the end and no suitable data found - signalize error
  r.f&=~F_CARRY;
  set_registers(&r);
}

